#include "rasterizer.h"
#include "graphics.h"
#include <stdlib.h>

rasterizer* rasterizer_init(long edge_data_size) {
  rasterizer *self = malloc(sizeof(rasterizer));
  
  self->edge_buffer = malloc(edge_data_size);
  self->edge_buffer_top = self->edge_buffer;
  self->top_edge = NULL;
  
  self->curve_stack_top = self->curve_stack;
  self->top_curve = NULL;
  
  self->current_point = make_point(0, 0);
  self->current_path_start = make_point(0, 0);
  
  self->inactive_edges = NULL;
  self->left_edges = NULL;
  self->right_edges = NULL;
  
  return self;
}

void reset(rasterizer *self) {
  // cleanup
  self->edge_buffer_top = self->edge_buffer;
  self->top_edge = NULL;
  
  self->curve_stack_top = self->curve_stack;
  self->top_curve = NULL;
  
  self->current_point = make_point(0, 0);
  self->current_path_start = make_point(0, 0);
  
  self->inactive_edges = NULL;
  self->left_edges = NULL;
  self->right_edges = NULL;
}

int path(rasterizer *self, control_point *points) {
  // check that first point is on curve
  if(points == NULL || !points->on_curve)
    return -1;
  
  // printf("Starting path at (%ld, %ld)\n", points->pos.x >> FBITS, points->pos.y >> FBITS);
  move_to(self, points->pos);
  
  control_point *cp = points->next;
  
  while(cp != NULL) {
    if(cp->on_curve) { // line
      line_to(self, cp->pos);
      cp = cp->next;
    } else { // curve
      if(!cp->next || cp->next->on_curve || !cp->next->next || !cp->next->next->on_curve) {
        printf("Bad path!\n");
        reset(self);
        return -1;
      }
      
      curve_to(self, cp->pos, cp->next->pos, cp->next->next->pos);
      
      if(cp->next->next == points) // closed curve
        break;
      
      cp = cp->next->next->next;
    }
  } 
  
  return 0;
}

void render(rasterizer *self, SDL_Surface *screen, Uint32 color) {
  long i;
  edge *left_edge, *right_edge;
  
  if(self->current_point.y != self->current_path_start.y)
    line_to(self, self->current_path_start); // close an open path
  
  /*
  printf("Rendering edges:\n");
  left_edge = self->edge_buffer;
  while(left_edge != NULL) {
    printf("Edge y_min: %ld, y_max: %ld, dir: %d, pixels: %ld\n", 
      left_edge->y_min, left_edge->y_max, left_edge->direction, left_edge->pixels_length);
    left_edge = left_edge->next;
  }
  left_edge = NULL;
  //*/
  
  // set inactive edges to the start of the edge buffer
  self->inactive_edges = self->edge_buffer;
  
  // sort inactive edges by top x coordinate
  sort_inactive_edges(self);
  
  // for each scanline 
  for(i = 0; i < screen->h && (self->left_edges || self->right_edges || self->inactive_edges) ; i++) {
    // initialize the scanline
    init_scanline(self, i);
    
    // go through the lists and draw segment from left to right edges, skip repeating values
    left_edge = self->left_edges;
    right_edge = self->right_edges;
      
    // printf("Drawing edges\n");
    
    while(left_edge != NULL && right_edge != NULL) {
      // printf("Drawing segment on y = %ld, between x1 = %ld and x2 = %ld\n", i, left_edge->pixels[i - left_edge->y_min], right_edge->pixels[right_edge->y_max - i]);    
      show_segment(screen, left_edge->pixels[i - left_edge->y_min], right_edge->pixels[right_edge->y_max - i], i, color);
      left_edge = left_edge->next;
      right_edge = right_edge->next;
    }
  }
  
  reset(self);
}

void move_to(rasterizer *self, point p0) {
  // printf("move_to: cp.y = %ld, cps.y = %ld\n", self->current_point.y, self->current_path_start.y);
  if(self->current_point.y != self->current_path_start.y)
    line_to(self, self->current_path_start);
  
  self->current_point = p0;
  self->current_path_start = p0;
}

void curve_to(rasterizer *self, point c0, point c1, point p1) {  
  
  /*
  if(c1.y < p1.y)
    p1.y++; // make the curve miss the scanline at the end (we don't want the last point yet)
  //*
  if(c0.y < self->current_point.y)
    self->current_point.y++; // make the curve cross (not only approach) the scanline at the end
  //*/
  
  // push the curve to the stack
  self->curve_stack[0] = self->current_point;
  self->curve_stack[1] = c0;
  self->curve_stack[2] = c1;
  self->curve_stack[3] = p1;
  self->top_curve = self->curve_stack;
  self->curve_stack_top = self->top_curve + 4; 
  
  rasterize_curves(self);
  
  self->current_point = p1;
}

void line_to(rasterizer *self, point p1) {
  point p0 = self->current_point;
  
  if(p0.y == p1.y) {
    self->current_point = p1;
    return;
  }
  
  long x1 = p0.x >> FBITS;
  long x2 = p1.x >> FBITS;
  long y1 = p0.y >> FBITS;
  long y2 = p1.y >> FBITS;
  
  long dx = abs(x2-x1);
  long dy;
  long e = 0, dif;
  
  if(y2 > y1) {
    start_edge(self, p0.y, p1.y + 1, dir_down);
    dy = y2 - y1;
    
    if(x2 > x1)
      dif = -1;
    else
      dif = 1;

    for(; y2 >= y1; y2--) {
      e -= dx;
      
      // add an edge pixel
      *((long *)self->edge_buffer_top) = x2;
      self->edge_buffer_top += sizeof(long);
      self->top_edge->pixels_length++;      

      while((e << 2) <= -dy && e != 0) {
        e += dy; 
        x2 += dif;
      }
    }
  } else {
    start_edge(self, p0.y + 1, p1.y, dir_up);
    dy = y1 - y2;
    
    if(x2 > x1)
      dif = -1;
    else
      dif = 1;

    for(; y2 <= y1; y2++) {
      e -= dx;
      
      // add an edge pixel
      *((long *)self->edge_buffer_top) = x2;
      self->edge_buffer_top += sizeof(long);
      self->top_edge->pixels_length++;      

      while((e << 2) <= -dy && e != 0) {
        e += dy; 
        x2 += dif;
      }
    }
  }
  
  self->current_point = p1;
}

// private API
void rasterize_curves(rasterizer *self) {  
  if(!self->top_curve)
    return;
  
  int new_curve = 1;
  monotony dir = curve_monotony(self->top_curve[0], self->top_curve[1], self->top_curve[2], self->top_curve[3]);

  // until the stack is empty, subdivide the curve and rasterize it 
  while(self->curve_stack_top > self->curve_stack + 1) {
    // printf("\nCurve step, %d curves on stack:\n", (int) (self->curve_stack_top - self->curve_stack) / 3);
    /*
    point *my_curve = self->curve_stack;
    while(my_curve < self->curve_stack_top - 1) {
      printf("Curve: (%ld, %ld), (%ld, %ld), (%ld, %ld), (%ld, %ld), w: %d, h: %d\n",
        my_curve[0].x >> FBITS, my_curve[0].y >> FBITS,
        my_curve[1].x >> FBITS, my_curve[1].y >> FBITS,
        my_curve[2].x >> FBITS, my_curve[2].y >> FBITS,
        my_curve[3].x >> FBITS, my_curve[3].y >> FBITS,
        abs((my_curve[3].x >> FBITS) - (my_curve[0].x >> FBITS)), 
        abs((my_curve[3].y >> FBITS) - (my_curve[0].y >> FBITS)) 
      );
      my_curve += 3;
    } 
    //*/
  
    // check the topmost curve monotony
    dir = curve_monotony(self->top_curve[0], self->top_curve[1], self->top_curve[2], self->top_curve[3]);
    if(dir == dir_none) { // not strictly monotonic
      // printf("Monotony unknown!\n");
      split_curve(self); // subdivide
    } else { 
      if(dir == dir_flat) { // flat (or too short to measure)
        // printf("Flat curve, dropping\n");
  
        // remove the curve from the stack
        self->top_curve -= 3;
        self->curve_stack_top -= 3;
      } else { // strictly monotonic:
        // printf("Strict monotony! %d\n", dir);
  
        int cc = sl_cross_count(self->top_curve);
        
        if(cc < 1) {
          // printf("Dropping curve\n");
          // remove the curve from the stack
          self->top_curve -= 3;
          self->curve_stack_top -= 3;
          continue;
        }
  
        if(cc > 1 || too_wide(self->top_curve)) {
          // printf("Splitting...\n");
          // subdivide the curve

          split_curve(self);
        } else { // == 1
          if(cc == 1) { 
            if(!self->top_edge || self->top_edge->direction != dir || new_curve == 1) {
              // start new edge if direction
              new_curve = 0;
              start_edge(self, self->top_curve[0].y, self->top_curve[3].y, dir);
            }
            
            /* 
            printf("Adding %ld-th pixel (x=%ld ~ %ld, y ~ %ld, y0=%ld ~ %ld, y3=%ld ~ %ld, y)\n", 
                                                  self->top_edge->pixels_length, 
                                                  self->top_curve[0].x, 
                                                  self->top_curve[0].x >> FBITS,
                                                  self->top_edge->direction == dir_up ? 
                                                    self->top_edge->y_min + self->top_edge->pixels_length : 
                                                    self->top_edge->y_max - self->top_edge->pixels_length,
                                                  self->top_curve[0].y, 
                                                  self->top_curve[0].y >> FBITS,
                                                  self->top_curve[3].y, 
                                                  self->top_curve[3].y >> FBITS
                                                );
            //*/
  
            if(((self->top_curve[0].y) >> FBITS) <= self->top_edge->y_min) {
              // printf("Updating y_min to: %ld\n", self->top_curve[0].y >> FBITS);
              self->top_edge->y_min = self->top_curve[0].y >> FBITS;
            }

            if(((self->top_curve[3].y) >> FBITS) <= self->top_edge->y_min) {
              // printf("Updating y_min to: %ld\n", self->top_curve[0].y >> FBITS);
              self->top_edge->y_min = self->top_curve[3].y >> FBITS;
            }
  
            if(((self->top_curve[0].y) >> FBITS) >= self->top_edge->y_max) {
              // printf("Updating y_max to: %ld\n", self->top_curve[0].y >> FBITS);
              self->top_edge->y_max = self->top_curve[0].y >> FBITS;
            }

            if(((self->top_curve[3].y) >> FBITS) >= self->top_edge->y_max) {
              // printf("Updating y_max to: %ld\n", self->top_curve[0].y >> FBITS);
              self->top_edge->y_max = self->top_curve[3].y >> FBITS;
            }
  
            // add an edge pixel
            *((long *)self->edge_buffer_top) = self->top_curve[0].x >> FBITS;
            self->edge_buffer_top += sizeof(long);
            self->top_edge->pixels_length++;
          }
  
          // printf("Dropping curve\n");
          // remove the curve from the stack
          self->top_curve -= 3;
          self->curve_stack_top -= 3; 
        }
      }
    }
  } // end until
  
  self->top_curve = NULL;
}

// private API
void split_curve(rasterizer *self) {
  long a, b, c;
  
  // x coordinate  
  a = self->top_curve[0].x + self->top_curve[1].x + 1;
  b = self->top_curve[1].x + self->top_curve[2].x + 1;
  c = self->top_curve[2].x + self->top_curve[3].x + 1;
  
  self->top_curve[6].x = self->top_curve[3].x;
  
  self->top_curve[1].x = a = a >> 1;
  b = b >> 1;
  self->top_curve[5].x = c = c >> 1;
  
  self->top_curve[2].x = a = (a + b + 1) >> 1;
  self->top_curve[4].x = b = (b + c + 1) >> 1;
  
  self->top_curve[3].x = (a + b + 1) >> 1;
  
  // same for y coordinate
  a = self->top_curve[0].y + self->top_curve[1].y + 1;
  b = self->top_curve[1].y + self->top_curve[2].y + 1;
  c = self->top_curve[2].y + self->top_curve[3].y + 1;
  
  self->top_curve[6].y = self->top_curve[3].y;
  
  self->top_curve[1].y = a = a >> 1;
  b = b >> 1;
  self->top_curve[5].y = c = c >> 1;
  
  self->top_curve[2].y = a = (a + b + 1) >> 1;
  self->top_curve[4].y = b = (b + c + 1) >> 1;
  
  self->top_curve[3].y = (a + b + 1) >> 1;
  
  // update stack pointers
  self->top_curve += 3;
  self->curve_stack_top += 3;
}

// private API
void start_edge(rasterizer *self, ffloat y1, ffloat y2, monotony direction) {
  // create an edge
  edge new_edge;
  
  // set minimum and maximum to the oposite end of a curve, so that they are out of bounds 
  // and will get updated correctly
  if(direction == dir_up) {
    new_edge.y_min = (y1 >> FBITS) + 1;
    new_edge.y_max = (y2 >> FBITS) - 1;
  } else {
    new_edge.y_min = (y2 >> FBITS) + 1;
    new_edge.y_max = (y1 >> FBITS) - 1;
  }
  
  // printf("Starting new edge (y_min: %ld, y_max: %ld)\n", new_edge.y_min, new_edge.y_max);
  
  new_edge.direction = direction;
  new_edge.pixels = (long *) (self->edge_buffer_top + sizeof(edge));
  new_edge.pixels_length = 0;
  new_edge.next = NULL;
  
  if(self->top_edge != NULL) {
    self->top_edge->next = (struct _edge *) self->edge_buffer_top; // link the last edge to the new
  }
  
  *((edge *) self->edge_buffer_top) = new_edge; // copy the edge
  self->top_edge = (edge*)self->edge_buffer_top; // start new edge
  self->edge_buffer_top += sizeof(edge); // move top pointer
}

// private API
void init_scanline(rasterizer *self, long scanline) {
  //printf("\nStarting new scanline: %ld\n", scanline);

  drop_finished_edges(self, scanline);
  sort_edge_lists(self, scanline);
  
  // search through inactive_edges for edges starting at scanline and insert them to left and right lists
  edge *inactive_edge = self->inactive_edges;
  edge *prev_inactive_edge = NULL;
  edge *left_edge = self->left_edges;
  edge *right_edge = self->right_edges;
  
  while(inactive_edge != NULL) { // for each inactive edge
    // skip edges that don't start at this scanline
    if(inactive_edge->y_min != scanline) {
      prev_inactive_edge = inactive_edge;
      inactive_edge = inactive_edge->next;
      continue;
    }
    
    // printf("Found inactive edge for insertion\n");

    if(inactive_edge->direction == dir_up) { // left edge
      while(left_edge != NULL) {
        if(!left_edge->next || left_edge->next->pixels[scanline - left_edge->next->y_min] > inactive_edge->pixels[0])
          break;
        
        left_edge = left_edge->next;
      }

      // remove inactive edge from its list
      if(prev_inactive_edge)
        prev_inactive_edge->next = inactive_edge->next;
      else
        self->inactive_edges = inactive_edge->next;
      
      /*
      if(left_edge) {
        printf("Left edge insertion point x: %ld\n", left_edge->pixels[scanline - left_edge->y_min]);
      }
      //*/
      // insert inactive_edge as new left edge
      if(left_edge && left_edge->pixels[scanline - left_edge->y_min] <= inactive_edge->pixels[0]) {
        // printf("Inserting inactive edge (x: %ld) as new left (after left x: %ld)\n", inactive_edge->pixels[0], left_edge->pixels[scanline - left_edge->y_min]);
        inactive_edge->next = left_edge->next;
        left_edge->next = inactive_edge;
      } else {
        // printf("Inserting inactive edge (x: %ld) as new left (first)\n", inactive_edge->pixels[0]);
        inactive_edge->next = self->left_edges;
        self->left_edges = inactive_edge;
        left_edge = self->left_edges;
      }
      
      if(prev_inactive_edge)
        inactive_edge = prev_inactive_edge->next;
      else
        inactive_edge = self->inactive_edges;
        
    } else { // right edge
      while(right_edge != NULL) {
        if(!right_edge->next || right_edge->next->pixels[right_edge->next->y_max - scanline] > inactive_edge->pixels[inactive_edge->pixels_length-1])
          break;
        
        right_edge = right_edge->next;
      }

      // remove inactive edge from its list
      if(prev_inactive_edge)
        prev_inactive_edge->next = inactive_edge->next;
      else
        self->inactive_edges = inactive_edge->next;
         
      // insert inactive_edge as new right edge
      if(right_edge && right_edge->pixels[right_edge->y_max - scanline] <= inactive_edge->pixels[inactive_edge->pixels_length-1]) {
        inactive_edge->next = right_edge->next;
        right_edge->next = inactive_edge;
      } else {
        inactive_edge->next = self->right_edges;
        self->right_edges = inactive_edge;
        right_edge = self->right_edges;
      }

      if(prev_inactive_edge)
        inactive_edge = prev_inactive_edge->next;
      else
        inactive_edge = self->inactive_edges;
    }
  }
  
  /*
  printf("Scanline initialization complete.\n");

  edge *an_edge;
  
  printf("Inactive edges:\n");
  an_edge = self->inactive_edges;
  while(an_edge != NULL) {
    printf("Edge y_min: %ld, y_max: %ld, dir: %s, pixels: %ld, top x: %ld\n", 
      an_edge->y_min, an_edge->y_max, (an_edge->direction == dir_up ? "up" : "down"), an_edge->pixels_length, 
      an_edge->pixels[(an_edge->direction == dir_up ? 0 : an_edge->pixels_length-1)]);
    an_edge = an_edge->next;
  }

  printf("Left edges:\n");
  an_edge = self->left_edges;
  while(an_edge != NULL) {
    printf("Edge y_min: %ld, y_max: %ld, dir: %d, pixels: %ld, x on this scanline: %ld\n", 
      an_edge->y_min, an_edge->y_max, an_edge->direction, an_edge->pixels_length, 
      an_edge->pixels[scanline - an_edge->y_min]);
    an_edge = an_edge->next;
  }

  printf("Right edges:\n");
  an_edge = self->right_edges;
  while(an_edge != NULL) {
    printf("Edge y_min: %ld, y_max: %ld, dir: %d, pixels: %ld, x on this scanline: %ld\n", 
      an_edge->y_min, an_edge->y_max, an_edge->direction, an_edge->pixels_length, 
      an_edge->pixels[an_edge->y_max - scanline]);
    an_edge = an_edge->next;
  }

  //*/
  
}

// private API
void drop_finished_edges(rasterizer *self, long scanline) {
  edge *an_edge;
  edge *prev_edge;
  
  // drop finished left edges
  prev_edge = NULL;
  an_edge = self->left_edges;
  while(an_edge != NULL) {
    if(an_edge->y_max < scanline) {
      if(prev_edge) {
        prev_edge->next = an_edge->next;
      } else {
        self->left_edges = an_edge->next;
      }
    } else {
      prev_edge = an_edge;
    }
    an_edge = an_edge->next;
  }

  // drop finished right edges
  prev_edge = NULL;
  an_edge = self->right_edges;
  while(an_edge != NULL) {
    if(an_edge->y_max < scanline) {
      if(prev_edge) {
        prev_edge->next = an_edge->next;
      } else {
        self->right_edges = an_edge->next;
      }
    } else {
      prev_edge = an_edge;
    }
    an_edge = an_edge->next;
  }
  
  // drop finished inactive edges
  prev_edge = NULL;
  an_edge = self->inactive_edges;
  while(an_edge != NULL) {
    if(an_edge->y_max < scanline) {
      if(prev_edge) {
        prev_edge->next = an_edge->next;
      } else {
        self->inactive_edges = an_edge->next;
      }
    } else {
      prev_edge = an_edge;
    }
    an_edge = an_edge->next;
  }
}

// private API
void sort_inactive_edges(rasterizer *self) {
  edge *cursor = self->inactive_edges;
  edge *prev = NULL;
  int ordered = 0;
  
  while(!ordered) {
    // printf("\ninactive_edges bubblesort iteration\n");
    ordered = 1;
    cursor = self->inactive_edges; // rewind
    prev = NULL;
    
    /*
    printf("Inactive edges list:\n");
    while(cursor != NULL) {
      printf("- Inactive edge (%ld) pixels: (%ld), y_min: %ld, y_max: %ld, dir: %d, top x: %ld\n",
        (long) cursor, cursor->pixels_length,
        cursor->y_min, cursor->y_max, cursor->direction,
        cursor->pixels[(cursor->direction == dir_up ? 0 : cursor->pixels_length - 1)]
        );
      
      if(cursor->pixels_length < 1) {
        exit(1);
      }
      
      cursor = cursor->next;
    }
    cursor = self->inactive_edges;
    //*/
    
    while(cursor->next != NULL) {
      if((cursor->direction == dir_up ? cursor->pixels[0] : cursor->pixels[cursor->pixels_length - 1]) >
         (cursor->next->direction == dir_up ? cursor->next->pixels[0] : cursor->next->pixels[cursor->next->pixels_length - 1])) {
        // swap
        if(prev != NULL) {
          // printf("Swapping non-first edge (%ld) with it's follower\n", (long) cursor);
          
          prev->next = cursor->next;
          cursor->next = cursor->next->next;
          prev->next->next = cursor;
          
          prev = prev->next;
        } else {
          // printf("Swapping first edge (%ld) with it's follower\n", (long) cursor);
          
          self->inactive_edges = cursor->next;
          cursor->next = cursor->next->next;
          self->inactive_edges->next = cursor;
          
          cursor = self->inactive_edges;
        }
        
        ordered = 0;
      } else {
        prev = cursor;
        cursor = cursor->next;  
      }  
    }
  }
}

// private API
void sort_edge_lists(rasterizer *self, long scanline) {
  // bubble sort the left and right list. 
  // NOTE the reasoning behind the choice of bubble sort is that the order of the lists barely changes between scanlines
  // NOTE edge pixels are sorted against the curve direction (up: top to bottom, down: bottom to top)
  
  edge *cursor = self->left_edges;
  edge *prev = NULL;
  int ordered = 0;
  
  while(!ordered && self->left_edges != NULL) {
    // printf("scanline initalization left bubblesort iteration\n");
    ordered = 1;
    cursor = self->left_edges; // rewind
    prev = NULL;
    
    /*
    printf("Left edges list:\n");
    while(cursor != NULL) {
      printf("- Left edge (%ld) pixels: (%ld), y_min: %ld, y_max: %ld, dir: %d, scanline x: %ld\n",
        (long) cursor, cursor->pixels_length,
        cursor->y_min, cursor->y_max, cursor->direction,
        cursor->pixels[scanline - cursor->y_min]
        );
      
      if(cursor->pixels_length < 1) {
        exit(1);
      }
      
      cursor = cursor->next;
    }
    cursor = self->left_edges;
    //*/
    
    while(cursor->next != NULL) {
      /* 
      printf("Left edge y_min: %ld, y_max: %ld, dir: %d, first x: %ld, last x: %ld\n",
        cursor->y_min, cursor->y_max, cursor->direction,
        cursor->pixels[0], cursor->pixels[cursor->pixels_length - 1]
        );
        printf("(Followed by edge y_min: %ld, y_max: %ld, dir: %d, first x: %ld, last x: %ld)\n",
        cursor->next->y_min, cursor->next->y_max, cursor->next->direction,
        cursor->next->pixels[0], cursor->next->pixels[cursor->next->pixels_length - 1]
        );
      //*/
      if(cursor->pixels[scanline - cursor->y_min] > cursor->next->pixels[scanline - cursor->next->y_min]) {
        // swap
        if(prev != NULL) {
          prev->next = cursor->next;
          cursor->next = cursor->next->next;
          prev->next->next = cursor;
          
          cursor = prev->next;
        } else {
          self->left_edges = cursor->next;
          cursor->next = cursor->next->next;
          self->left_edges->next = cursor;
          
          cursor = self->left_edges;
          prev = NULL;
        }
        
        ordered = 0;
      }
      
      prev = cursor;      
      cursor = cursor->next;
    }
  }
  
  cursor = self->right_edges;
  prev = NULL;
  ordered = 0;
  
  while(!ordered && self->right_edges != NULL) {
    // printf("scanline %ld initalization right bubblesort iteration\n", scanline);
    ordered = 1;
    cursor = self->right_edges; // rewind
    prev = NULL;
    
    /*
    printf("Right edges list:\n");
    while(cursor != NULL) {
      printf("- Right edge (%ld) pixels: (%ld), y_min: %ld, y_max: %ld, dir: %d, scanline x: %ld\n",
        (long) cursor, cursor->pixels_length,
        cursor->y_min, cursor->y_max, cursor->direction,
        cursor->pixels[cursor->y_max - scanline]
        );
      
      if(cursor->pixels_length < 1) {
        exit(1);
      }
      
      cursor = cursor->next;
    }
    cursor = self->right_edges;
    //*/
    
    while(cursor->next != NULL) {
      /* 
      printf("Right edge (%ld) y_min: %ld, y_max: %ld, dir: %d, cur x: %ld\n",
        (long) cursor,
        cursor->y_min, cursor->y_max, cursor->direction,
        cursor->pixels[cursor->y_max - scanline]
        );
      printf("(Followed by edge (%ld) y_min: %ld, y_max: %ld, dir: %d, cur x: %ld)\n",
        (long) cursor->next,
        cursor->next->y_min, cursor->next->y_max, cursor->next->direction,
        cursor->next->pixels[cursor->next->y_max - scanline]
        );
      //*/
      if(cursor->pixels[cursor->y_max - scanline] > cursor->next->pixels[cursor->next->y_max - scanline]) {
        // swap
        if(prev != NULL) {
          prev->next = cursor->next;
          cursor->next = cursor->next->next;
          prev->next->next = cursor;
          
          cursor = prev->next;
        } else {
          self->right_edges = cursor->next;
          cursor->next = cursor->next->next;
          self->right_edges->next = cursor;
          
          cursor = self->right_edges;
          prev = NULL;
        }
        
        ordered = 0;
      }

      prev = cursor;
      cursor = cursor->next;
    }
  }
  
}

// private API
monotony curve_monotony(point p0, point c0, point c1, point p1) {
  /*printf("Monotony? (%ld.%ld, %ld.%ld, %ld.%ld, %ld.%ld)\n", 
    p0.x, p0.y,
    c0.x, c0.y,
    c1.x, c1.y,
    p1.x, p1.y);
    */
  if(p0.y == c0.y && p0.y == c1.y && p0.y == p1.y) // special flat case
    return dir_flat;
  
  if(p0.y > p1.y && (c0.y <= p0.y && c0.y >= p1.y) && (c1.y <= p0.y && c1.y >= p1.y))
    return dir_up;
  
  if(p0.y < p1.y && (c0.y >= p0.y && c0.y <= p1.y) && (c1.y >= p0.y && c1.y <= p1.y))
    return dir_down;
  
  return dir_none;
}

// private API
int too_wide(point *curve) {
  // printf("Width check: %ld ~ %ld = %ld ~ %ld\n",
  //   curve[0].x,
  //   curve[0].x >> FBITS,
  //   curve[3].x,
  //   curve[3].x >> FBITS
  // );
  if(abs(curve[3].x - curve[0].x) <= 3) // handles precision problem
    return 0;
  
  return (curve[0].x >> FBITS) != (curve[3].x >> FBITS);
}

// private API
int sl_cross_count(point *curve) {
  unsigned long y0, y1, sl;
  
  if(curve[0].y < curve[3].y) {
    y0 = curve[0].y;
    y1 = curve[3].y;
  } else {
    y0 = curve[3].y;
    y1 = curve[0].y;
  }
  
  sl = (ffloat) ((y0 >> FBITS) << FBITS) + 32;
  int count = 1;
  // printf("Scanline intersection count (sl: %ld, sl2: %ld, y1: %ld ~ %ld, y2: %ld ~ %ld) is ", 
  //    sl,
  //    sl + 64,
  //    y0,
  //    y0 >> FBITS,
  //    y1,
  //    y1 >> FBITS
  //  );

  if(y1 >> FBITS > y0 >> FBITS) // curve must fit within a scanline
    count = 2;
  
  if(y0 > sl || y1 <= sl) // curve must cros a scanline centerline
    count = 0;
  
  if(y1 >= sl + 64)
    count = 2;  
  
  // printf("%d\n", count);
  
  return count;
}